from supervisely.app.widgets import Text, Select, Field, Container, Button
from supervisely.imaging.color import hex2rgb

device = Select(
    items=[
        Select.Item(label="CPU", value="cpu"),
        Select.Item(label="GPU 0", value="cuda:0"),
        Select.Item(label="GPU 1", value="cuda:1"),
        Select.Item(label="GPU 2", value="cuda:2"),
        Select.Item(label="GPU 3", value="cuda:3"),
    ],
    filterable=True,
)

deploy_btn = Button("Deploy model on device", icon="zmdi zmdi-flash")
hide_btn = Button("Hide")
show_btn = Button("Show")


@hide_btn.click
def hide():
    device.enable()


@show_btn.click
def show():
    device.disable()


# icon = Field.Icon(
#     zmdi_class="zmdi zmdi-memory",
#     # color_rgb=hex2rgb("#4977ff"),
#     # bg_color_rgb=hex2rgb("#ddf2ff"),
#     # color_rgb=hex2rgb("#2cd26e"),
#     # bg_color_rgb=hex2rgb("#d8f8e7"),
#     bg_color_rgb=hex2rgb("#2cd26e"),
# )
device_field = Field(
    content=device,
    title="Select device",
    description="Model will be loaded (deployed) on selected device: CPU or GPU",
)

# how to hide widget?
deploy_status = Text("Model has been successfully deployed", "success")

deploy_layout = Container([device_field, deploy_btn, deploy_status, hide_btn, show_btn])



import os
from typing import List, Dict
import yaml
from supervisely._utils import (
    is_production,
    is_development,
    is_debug_with_sly_net,
    rand_str,
)
from supervisely.app.fastapi.subapp import get_name_from_env
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag_meta import TagMeta, TagValueType

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
import supervisely.imaging.image as sly_image
import supervisely.io.fs as fs
from supervisely.sly_logger import logger

from supervisely.project.project_meta import ProjectMeta
from supervisely.app.fastapi.subapp import Application
from supervisely.app.content import StateJson, get_data_dir
from supervisely.app.fastapi.request import Request
from supervisely.api.api import Api
import supervisely.app.development as sly_app_development
from distinctipy import distinctipy

try:
    from typing import Literal
except ImportError:
    # for compatibility with python 3.7
    from typing_extensions import Literal


class Inference:
    def __init__(
        self,
        model_dir: str = None,
        device: Literal["cpu", "cuda", "cuda:0", "cuda:1", "cuda:2", "cuda:3"] = "cpu",
    ):
        self._model_dir = model_dir
        self._device = device
        self._model_meta = None
        self._confidence = "confidence"
        self._app: Application = None
        self._api: Api = None

        self._headless = False
        # self._template_dir = None
        # self._template_dir = Path(__file__).parent.absolute()
        if is_production():
            if os.environ.get("_SPAWN_USER_ID") is None:
                logger.debug("Running serving on localhost with enabled UI")
            else:
                logger.debug(
                    "Running serving on Supervisely platform in production mode"
                )
                raise NotImplementedError("TBD - download directory")
        elif is_development():
            self._headless = True
            pass

    def _get_templates_dir(self):
        raise NotImplementedError("Have to be implemented in child class")

    def load_on_device(
        device: Literal["cpu", "cuda", "cuda:0", "cuda:1", "cuda:2", "cuda:3"] = "cpu"
    ):
        raise NotImplementedError(
            "Have to be implemented in child class after inheritance"
        )

    def get_classes(self) -> List[str]:
        raise NotImplementedError(
            "Have to be implemented in child class after inheritance"
        )

    def get_info(self) -> dict:
        return {
            "model name": get_name_from_env(default="Neural Network Serving"),
            "model dir": self.model_dir,
            "number of classes": len(self.get_classes()),
        }

    @property
    def model_dir(self):
        return self._model_dir

    @property
    def api(self) -> Api:
        if self._api is None:
            self._api = Api()
        return self._api

    def _get_obj_class_shape(self):
        raise NotImplementedError("Have to be implemented in child class")

    @property
    def model_meta(self) -> ProjectMeta:
        if self._model_meta is None:
            colors = distinctipy.get_colors(len(self.get_classes()))
            classes = []
            for name, color in zip(self.get_classes(), colors):
                rgb = distinctipy.get_rgb256(color)
                classes.append(ObjClass(name, self._get_obj_class_shape(), rgb))
            self._model_meta = ProjectMeta(classes)
            self._get_confidence_tag_meta()
        return self._model_meta

    def _get_confidence_tag_meta(self):
        tag_meta = self.model_meta.get_tag_meta(self._confidence)
        if tag_meta is None:
            tag_meta = TagMeta(self._confidence, TagValueType.ANY_NUMBER)
            self._model_meta = self._model_meta.add_tag_meta(tag_meta)
        return tag_meta

    def _create_label(self, dto) -> Label:
        raise NotImplementedError("Have to be implemented in child class")

    def predict(self, image_path: str):
        raise NotImplementedError("Have to be implemented in child class")

    def predict_annotation(self, image_path: str) -> Annotation:
        raise NotImplementedError("Have to be implemented in child class")

    def visualize(self, predictions: List, image_path: str, vis_path: str):
        raise NotImplementedError("Have to be implemented in child class")

    def _predictions_to_annotation(
        self, image_path: str, predictions: List
    ) -> Annotation:
        labels = []
        for prediction in predictions:
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = Annotation.from_img_path(image_path)
        ann = ann.add_labels(labels)
        return ann

    def _get_custom_inference_settings() -> str:  # in yaml format
        return ""

    def inference_image_path(
        self, image_path, project_meta: ProjectMeta, state: Dict, settings: Dict = None
    ):
        raise NotImplementedError()

    def _get_custom_inference_settings_dict(self) -> dict:
        return yaml.safe_load(self._get_custom_inference_settings())

    def get_inference_settings(self, state: dict):
        settings = state.get("settings", {})
        for key, value in self._get_custom_inference_settings_dict().items():
            if key not in settings:
                logger.warn(
                    f"Field {key} not found in inference settings. Use default value {value}"
                )
                settings[key] = value
        return settings

    @property
    def app(self) -> Application:
        return self._app

    def inference_batch_ids(self, api: Api, state: dict):
        ids = state["batch_ids"]
        infos = api.image.get_info_by_id_batch(ids)
        paths = []
        temp_dir = os.path.join(get_data_dir(), rand_str(10))
        fs.mkdir(temp_dir)
        for info in infos:
            paths.append(os.path.join(temp_dir, f"{rand_str(10)}_{info.name}"))
        api.image.download_paths(infos[0].dataset_id, ids, paths)
        results = self.inference_images_dir(paths, state)
        fs.remove_dir(temp_dir)
        return results

    def inference_images_dir(self, img_paths: List[str], state: Dict):
        settings = self.get_inference_settings(state)
        annotations = []
        for image_path in img_paths:
            ann_json = self.inference_image_path(
                image_path=image_path,
                project_meta=self.model_meta,
                state=state,
                settings=settings,
            )
            annotations.append(ann_json)
        return annotations

    def inference_image_id(self, api: Api, state: dict):
        logger.debug("Input state", extra={"state": state})
        settings = self.get_inference_settings(state)
        image_id = state["image_id"]
        image_info = api.image.get_info_by_id(image_id)
        image_path = os.path.join(get_data_dir(), f"{rand_str(10)}_{image_info.name}")
        api.image.download_path(image_id, image_path)
        ann_json = self.inference_image_path(
            image_path=image_path,
            project_meta=self.model_meta,
            state=state,
            settings=settings,
        )
        fs.silent_remove(image_path)
        return ann_json

    def inference_image_url(self, api: Api, state: dict):
        logger.debug("Input data", extra={"state": state})
        settings = self.get_inference_settings(state)
        image_url = state["image_url"]
        ext = fs.get_file_ext(image_url)
        if ext == "":
            ext = ".jpg"
        image_path = os.path.join(get_data_dir(), rand_str(15) + ext)
        fs.download(image_url, image_path)
        ann_json = self.inference_image_path(
            image_path=image_path,
            project_meta=self.model_meta,
            state=state,
            settings=settings,
        )
        fs.silent_remove(image_path)
        return ann_json

    def serve(self):
        if is_debug_with_sly_net():
            # advanced debug for Supervisely Team
            logger.warn(
                "Serving is running in advanced development mode with Supervisely VPN Network"
            )
            team_id = int(os.environ["context.teamId"])
            # sly_app_development.supervisely_vpn_network(action="down") # for debug
            sly_app_development.supervisely_vpn_network(action="up")
            task = sly_app_development.create_debug_task(team_id, port="8000")

        self._app = Application(
            headless=self._headless, templates_dir=self._get_templates_dir()
        )
        server = self._app.get_server()

        @server.post(f"/get_session_info")
        def get_session_info():
            return self.get_info()

        @server.post("/get_custom_inference_settings")
        def get_custom_inference_settings():
            settings = self._get_custom_inference_settings()
            return {"settings": settings}

        @server.post("/get_output_classes_and_tags")
        def get_output_classes_and_tags():
            return self.model_meta.to_json()

        @server.post("/inference_image_id")
        def inference_image_id(request: Request):
            return self.inference_image_id(request.api, request.state)

        @server.post("/inference_image_url")
        def inference_image_url(request: Request):
            return self.inference_image_url(request.api, request.state)

        @server.post("/inference_batch_ids")
        def inference_batch_ids(request: Request):
            return self.inference_batch_ids(request.api, request.state)

        @server.post("/inference_video_id")
        def inference_video_id(request: Request):
            raise NotImplementedError()



# coding: utf-8
"""api connection to the server which allows user to communicate with Supervisely"""

from __future__ import annotations
import os
import requests
import json
from typing import List, Optional, NamedTuple, Dict
from requests_toolbelt import MultipartEncoderMonitor, MultipartEncoder

import supervisely.api.team_api as team_api
import supervisely.api.workspace_api as workspace_api
import supervisely.api.project_api as project_api
import supervisely.api.neural_network_api as neural_network_api
import supervisely.api.task_api as task_api
import supervisely.api.dataset_api as dataset_api
import supervisely.api.image_api as image_api
import supervisely.api.annotation_api as annotation_api
import supervisely.api.plugin_api as plugin_api
import supervisely.api.agent_api as agent_api
import supervisely.api.role_api as role_api
import supervisely.api.user_api as user_api
import supervisely.api.labeling_job_api as labeling_job_api
import supervisely.api.video.video_api as video_api
import supervisely.api.pointcloud.pointcloud_api as pointcloud_api
import supervisely.api.pointcloud.pointcloud_episode_api as pointcloud_episode_api
import supervisely.api.object_class_api as object_class_api
import supervisely.api.report_api as report_api
import supervisely.api.app_api as app_api
import supervisely.api.file_api as file_api
import supervisely.api.image_annotation_tool_api as image_annotation_tool_api
import supervisely.api.advanced_api as advanced_api
import supervisely.api.import_storage_api as import_stoarge_api
import supervisely.api.remote_storage_api as remote_storage_api
import supervisely.api.github_api as github_api
import supervisely.api.volume.volume_api as volume_api
from supervisely.sly_logger import logger


from supervisely.io.network_exceptions import (
    process_requests_exception,
    process_unhandled_request,
)

SUPERVISELY_TASK_ID = "SUPERVISELY_TASK_ID"
SUPERVISELY_PUBLIC_API_RETRIES = "SUPERVISELY_PUBLIC_API_RETRIES"
SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC = "SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC"
SERVER_ADDRESS = "SERVER_ADDRESS"
API_TOKEN = "API_TOKEN"
TASK_ID = "TASK_ID"


class Api:
    """
    An API connection to the server with which you can communicate with your teams, workspaces and projects. :class:`Api<Api>` object is immutable.

    :param server_address: Address of the server.
    :type server_address: str
    :param token: Unique secret token associated with your agent.
    :type token: str
    :param retry_count: The number of attempts to connect to the server.
    :type retry_count: int, optional
    :param retry_sleep_sec: The number of seconds to delay between attempts to connect to the server.
    :type retry_sleep_sec: int, optional
    :param external_logger: Logger class object.
    :type external_logger: logger, optional
    :param ignore_task_id:
    :type ignore_task_id: bool, optional
    :raises: :class:`ValueError`, if token is None or it length != 128
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()
    """

    def __init__(
        self,
        server_address: str = None,
        token: str = None,
        retry_count: Optional[int] = None,
        retry_sleep_sec: Optional[int] = None,
        external_logger: Optional[logger] = None,
        ignore_task_id: Optional[bool] = False,
    ):
        if server_address is None and token is None:
            server_address = os.environ.get(SERVER_ADDRESS, None)
            token = os.environ.get(API_TOKEN, None)

        if server_address is None:
            raise ValueError(
                "SERVER_ADDRESS env variable is undefined, https://developer.supervise.ly/getting-started/basics-of-authentication"
            )
        if token is None:
            raise ValueError(
                "API_TOKEN env variable is undefined, https://developer.supervise.ly/getting-started/basics-of-authentication"
            )
        self.server_address = Api.normalize_server_address(server_address)

        if retry_count is None:
            retry_count = int(os.getenv(SUPERVISELY_PUBLIC_API_RETRIES, "20"))
        if retry_sleep_sec is None:
            retry_sleep_sec = int(
                os.getenv(SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC, "1")
            )

        if len(token) != 128:
            raise ValueError("Invalid token {!r}: length != 128".format(token))

        self.token = token
        self.headers = {"x-api-key": token}
        self.task_id = os.getenv(SUPERVISELY_TASK_ID)
        if self.task_id is not None and ignore_task_id is False:
            self.headers["x-task-id"] = self.task_id
        self.context = {}
        self.additional_fields = {}

        self.team = team_api.TeamApi(self)
        self.workspace = workspace_api.WorkspaceApi(self)
        self.project = project_api.ProjectApi(self)
        self.model = neural_network_api.NeuralNetworkApi(self)
        self.task = task_api.TaskApi(self)
        self.dataset = dataset_api.DatasetApi(self)
        self.image = image_api.ImageApi(self)
        self.annotation = annotation_api.AnnotationApi(self)
        self.plugin = plugin_api.PluginApi(self)
        self.agent = agent_api.AgentApi(self)
        self.role = role_api.RoleApi(self)
        self.user = user_api.UserApi(self)
        self.labeling_job = labeling_job_api.LabelingJobApi(self)
        self.video = video_api.VideoApi(self)
        # self.project_class = project_class_api.ProjectClassApi(self)
        self.object_class = object_class_api.ObjectClassApi(self)
        self.report = report_api.ReportApi(self)
        self.pointcloud = pointcloud_api.PointcloudApi(self)
        self.pointcloud_episode = pointcloud_episode_api.PointcloudEpisodeApi(self)
        self.app = app_api.AppApi(self)
        self.file = file_api.FileApi(self)
        self.img_ann_tool = image_annotation_tool_api.ImageAnnotationToolApi(self)
        self.advanced = advanced_api.AdvancedApi(self)
        self.import_storage = import_stoarge_api.ImportStorageApi(self)
        self.remote_storage = remote_storage_api.RemoteStorageApi(self)
        self.github = github_api.GithubApi(self)
        self.volume = volume_api.VolumeApi(self)

        self.retry_count = retry_count
        self.retry_sleep_sec = retry_sleep_sec

        self.logger = external_logger or logger

    @classmethod
    def normalize_server_address(cls, server_address):
        """ """
        result = server_address.strip("/")
        if ("http://" not in result) and ("https://" not in result):
            result = "http://" + result
        return result

    @classmethod
    def from_env(cls, retry_count: int = 5, ignore_task_id: bool = False) -> Api:
        """
        Initialize API use environment variables.

        :param retry_count: The number of attempts to connect to the server.
        :type retry_count: int
        :param ignore_task_id:
        :type ignore_task_id: bool
        :return: Api object
        :rtype: :class:`Api<supervisely.api.api.Api>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()
        """
        return cls(
            os.environ[SERVER_ADDRESS],
            os.environ[API_TOKEN],
            retry_count=retry_count,
            ignore_task_id=ignore_task_id,
        )

    def add_header(self, key: str, value: str) -> None:
        """
        Add given key and value to headers dictionary.

        :param key: New key.
        :type key: str
        :param value: New value.
        :type value: str
        :raises: :class:`RuntimeError`, if key is already set
        :return: None
        :rtype: :class:`NoneType`
        """
        if key in self.headers:
            raise RuntimeError(
                f"Header {key!r} is already set for the API object. "
                f"Current value: {self.headers[key]!r}. Tried to set value: {value!r}"
            )
        self.headers[key] = value

    def add_additional_field(self, key: str, value: str) -> None:
        """
        Add given key and value to additional_fields dictionary.

        :param key: New key.
        :type key: str
        :param value: New value.
        :type value: str
        :return: None
        :rtype: :class:`NoneType`
        """
        self.additional_fields[key] = value

    def post(
        self,
        method: str,
        data: Dict,
        retries: Optional[int] = None,
        stream: Optional[bool] = False,
    ) -> requests.Response:
        """
        Performs POST request to server with given parameters.

        :param method:
        :type method: str
        :param data: Dictionary to send in the body of the :class:`Request`.
        :type data: dict
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param stream: Define, if you’d like to get the raw socket response from the server.
        :type stream: bool, optional
        :return: Response object
        :rtype: :class:`Response<Response>`
        """
        if retries is None:
            retries = self.retry_count

        for retry_idx in range(retries):
            url = self.server_address + "/public/api/v3/" + method
            response = None
            try:
                if type(data) is bytes:
                    response = requests.post(
                        url, data=data, headers=self.headers, stream=stream
                    )
                elif (
                    type(data) is MultipartEncoderMonitor
                    or type(data) is MultipartEncoder
                ):
                    response = requests.post(
                        url,
                        data=data,
                        headers={**self.headers, "Content-Type": data.content_type},
                        stream=stream,
                    )
                else:
                    json_body = data
                    if type(data) is dict:
                        json_body = {**data, **self.additional_fields}
                    response = requests.post(
                        url, json=json_body, headers=self.headers, stream=stream
                    )

                if response.status_code != requests.codes.ok:
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(
                    self.logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=self.retry_sleep_sec,
                    response=response,
                    retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)
        raise requests.exceptions.RetryError("Retry limit exceeded ({!r})".format(url))

    def get(
        self,
        method: str,
        params: Dict,
        retries: Optional[int] = None,
        stream: Optional[bool] = False,
        use_public_api: Optional[bool] = True,
    ) -> requests.Response:
        """
        Performs GET request to server with given parameters.

        :param method:
        :type method: str
        :param params: Dictionary to send in the body of the :class:`Request`.
        :type method: dict
        :param retries: The number of attempts to connect to the server.
        :type method: int, optional
        :param stream: Define, if you’d like to get the raw socket response from the server.
        :type method: bool, optional
        :param use_public_api:
        :type method: bool, optional
        :return: Response object
        :rtype: :class:`Response<Response>`
        """
        if retries is None:
            retries = self.retry_count

        for retry_idx in range(retries):
            url = self.server_address + "/public/api/v3/" + method
            if use_public_api is False:
                url = os.path.join(self.server_address, method)
            response = None
            try:
                json_body = params
                if type(params) is dict:
                    json_body = {**params, **self.additional_fields}
                response = requests.get(
                    url, params=json_body, headers=self.headers, stream=stream
                )

                if response.status_code != requests.codes.ok:
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                process_requests_exception(
                    self.logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=self.retry_sleep_sec,
                    response=response,
                    retry_info={"retry_idx": retry_idx + 2, "retry_limit": retries},
                )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)

    @staticmethod
    def _raise_for_status(response):
        """
        Raise error and show message with error code if given response can not connect to server.
        :param response: Request class object
        """
        http_error_msg = ""
        if isinstance(response.reason, bytes):
            try:
                reason = response.reason.decode("utf-8")
            except UnicodeDecodeError:
                reason = response.reason.decode("iso-8859-1")
        else:
            reason = response.reason

        if 400 <= response.status_code < 500:
            http_error_msg = "%s Client Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                response.content.decode("utf-8"),
            )

        elif 500 <= response.status_code < 600:
            http_error_msg = "%s Server Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                response.content.decode("utf-8"),
            )

        if http_error_msg:
            raise requests.exceptions.HTTPError(http_error_msg, response=response)

    @staticmethod
    def parse_error(
        response: requests.Response,
        default_error: Optional[str] = "Error",
        default_message: Optional[str] = "please, contact administrator",
    ):
        """
        Processes error from response.

        :param response: Request object.
        :type method: Request
        :param default_error: Error description.
        :type method: str, optional
        :param default_message: Message to user.
        :type method: str, optional
        :return: Number of error and message about curren connection mistake
        :rtype: :class:`int`, :class:`str`
        """
        ERROR_FIELD = "error"
        MESSAGE_FIELD = "message"
        DETAILS_FIELD = "details"

        try:
            data_str = response.content.decode("utf-8")
            data = json.loads(data_str)
            error = data.get(ERROR_FIELD, default_error)
            details = data.get(DETAILS_FIELD, {})
            if type(details) is dict:
                message = details.get(MESSAGE_FIELD, default_message)
            else:
                message = details[0].get(MESSAGE_FIELD, default_message)

            return error, message
        except Exception as e:
            return "", ""

    def pop_header(self, key: str) -> str:
        """ """
        if key not in self.headers:
            raise KeyError(f"Header {key!r} not found")
        return self.headers.pop(key)



import os
from pathlib import Path
import shlex
import subprocess
from supervisely.io.fs import mkdir
from supervisely.api.api import Api
from supervisely.sly_logger import logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

VPN_CONFIGURATION_DIR = "~/supervisely-network"


def supervisely_vpn_network(action: Literal["up", "down"] = "up"):
    # TODO: already down "wg-quick: `wg0' is not a WireGuard interface\n"
    # TODO: wg-quick must be run as root. Please enter the password for max to continue:

    process = subprocess.run(
        shlex.split("curl --max-time 3 -s http://10.8.0.1:80"),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    if "Connected to Supervisely Net!" in process.stdout:
        logger.info(f"You connected to Supervisely VPN Network")
        return

    api = Api()
    current_dir = Path(__file__).parent.absolute()
    script_path = os.path.join(current_dir, "sly-net.sh")
    network_dir = os.path.expanduser(VPN_CONFIGURATION_DIR)
    mkdir(network_dir)

    process = subprocess.run(
        shlex.split(
            f"{script_path} {action} {api.token} {api.server_address} {network_dir}"
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    text = "connected to"
    if action == "down":
        text = "disconnected from"
    try:
        process.check_returncode()
        logger.info(f"You have been successfully {text} to Supervisely VPN Network")
    except subprocess.CalledProcessError as e:
        print(e.stdout)

        e.cmd[2] = "***-api-token-***"
        if "wg0' already exists" in e.stderr:
            logger.info(f"You {text} Supervisely VPN Network")
            pass
        else:
            raise e


def create_debug_task(team_id, port="8000"):
    api = Api()
    me = api.user.get_my_info()
    session_name = me.login + "-development"
    module_id = api.app.get_ecosystem_module_id(
        "supervisely-ecosystem/while-true-script-v2"
    )
    sessions = api.app.get_sessions(team_id, module_id, session_name=session_name)
    redirect_requests = {"token": api.token, "port": port}
    task = None
    for session in sessions:
        if session["meta"]["redirectRequests"] == redirect_requests:
            task = session
            logger.info(f"Debug task already exists: {task['id']}")
            break
    workspaces = api.workspace.get_list(team_id)
    if task is None:
        task = api.task.start(
            agent_id=None,
            module_id=module_id,
            workspace_id=workspaces[0].id,
            task_name=session_name,
            redirect_requests=redirect_requests,
            proxy_keep_url=False,  # to ignore /net/<token>/endpoint
        )
        if type(task) is list:
            task = task[0]
        logger.info(f"Debug task has been successfully created: {task['taskId']}")
    return task



# coding: utf-8

import queue
import logging
import threading
import concurrent.futures
import time
import json

from supervisely.sly_logger import get_task_logger, add_logger_handler, change_formatters_default_values, \
                                       ServiceType, EventType, add_default_logging_into_file, logger
from supervisely.api.module_api import ApiField

BATCH_SIZE_LOG = 50


class LogQueue:
    def __init__(self):
        self.q = queue.Queue()  # no limit

    def put_nowait(self, log_line):
        self.q.put_nowait(log_line)

    def _get_batch_nowait(self, batch_limit):
        log_lines = []
        for _ in range(batch_limit):
            try:
                log_line = self.q.get_nowait()
            except queue.Empty:
                break
            log_lines.append(log_line)
        return log_lines

    def get_log_batch_nowait(self):
        res = self._get_batch_nowait(BATCH_SIZE_LOG)
        return res

    def get_log_batch_blocking(self):
        first_log_line = self.q.get(block=True)
        rest_lines = self._get_batch_nowait(BATCH_SIZE_LOG)
        log_lines = [first_log_line] + rest_lines
        return log_lines


class SlyApiHandler(logging.Handler):
    def __init__(self, api):
        super().__init__()
        self._api = api
        self._stop_log_event = threading.Event()

        self.log_queue = LogQueue()
        self.executor_log = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self.future_log = self.executor_log.submit(self.submit_log)

    def emit(self, record):
        log_entry = self.format(record)
        log_entry = json.loads(log_entry)
        self.log_queue.put_nowait(log_entry)

    def submit_log(self):
        break_flag = False
        while True:
            log_lines = self.log_queue.get_log_batch_nowait()
            if len(log_lines) > 0:
                self._api.task.submit_logs(log_lines)
                break_flag = False
            else:
                if break_flag:
                    return True
                if self._stop_log_event.isSet():
                    print("stop here")
                    break_flag = True  # exit after next loop without data
                time.sleep(0.5)

    def stop_log_thread(self):
        self._stop_log_event.set()
        self.executor_log.shutdown(wait=True)
        return self.future_log.result()  # crash if log thread crashed


def init_global_task_logger(task_id, api=None, file_path=None):
    change_formatters_default_values(logger, 'service_type', ServiceType.TASK)
    change_formatters_default_values(logger, 'event_type', EventType.LOGJ)
    change_formatters_default_values(logger, 'task_id', task_id)

    if api is not None:
        handler = SlyApiHandler(api)
        add_logger_handler(logger, handler)
    if file_path is not None:
        #add_default_logging_into_file(logger, self.dir_logs)
        pass


def _stop_and_wait_logger(logger):
    for handler in logger.handlers:
        if type(handler) is SlyApiHandler:
            handler.stop_log_thread()


def log_task_finished(logger):
    if logger is None:
        return
    logger.info('TASK_END', extra={'event_type': EventType.TASK_FINISHED})
    _stop_and_wait_logger(logger)


def log_task_crashed(logger, e=None):
    if logger is None:
        return
    if e is None:
        e = Exception("Crashed without exception info")
    logger.critical('TASK_END', exc_info=True, extra={'event_type': EventType.TASK_CRASHED, 'exc_str': str(e)})
    _stop_and_wait_logger(logger)






from typing import Dict, List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import (
    Widget,
    SelectDataset,
    Checkbox,
    Select,
    Text,
    generate_id,
)
from supervisely.api.api import Api
from supervisely.app.widgets.select_sly_utils import _get_int_or_env


class SelectItem(Widget):
    def __init__(
        self,
        dataset_id: int = None,
        compact: bool = True,
        show_label: bool = True,
        size: Literal["large", "small", "mini"] = None,
        widget_id: str = None,
    ):
        self._api = Api()
        self._dataset_id = dataset_id
        self._compact = compact
        self._show_label = show_label
        self._size = size

        items = []
        if self._dataset_id is not None:
            items = self._get_items_from_server(dataset_id)
            need_limit, ds_items_count = self._need_set_limit(dataset_id)
            if need_limit is True:
                self._set_limit(ds_items_count)

        self._item_selector = Select(
            items=items, filterable=True, size=size, widget_id=generate_id()
        )
        self._ds_selector = None
        self._limit_message = None

        self._dataset_id = _get_int_or_env(self._dataset_id, "modal.state.slyDatasetId")
        if self._compact is True:
            if self._dataset_id is None:
                raise ValueError(
                    '"dataset_id" have to be passed as argument or "compact" has to be False'
                )
        else:
            self._show_label = True
            self._ds_selector = SelectDataset(
                default_id=self._dataset_id,
                multiselect=False,
                compact=self._compact,
                show_label=True,
                size=self._size,
                widget_id=generate_id(),
            )

            @self._ds_selector.value_changed
            def dataset_changed(dataset_id):
                # print("dataset_id ---> ", dataset_id)
                self.refresh_items(dataset_id)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        res = {
            "limit_message": self._limit_message,
            "compact": self._compact,
            "show_label": self._show_label,
        }
        return res

    def get_json_state(self) -> Dict:
        res = {}
        return res

    def get_selected_id(self):
        return self._item_selector.get_value()

    def _need_set_limit(self, dataset_id: int = None, limit: int = 50):
        count = None
        if dataset_id is not None:
            dataset_info = self._api.dataset.get_info_by_id(dataset_id)
            count = dataset_info.items_count
            if count > limit:
                return True, count
        return False, count

    def _set_limit(self, dataset_items: int = None, limit: int = 50):
        self._limit_message = (
            f"Showed first {limit} of {dataset_items} items in dataset"
        )
        DataJson()[self.widget_id]["limit_message"] = self._limit_message
        DataJson().send_changes()

    def _get_items_from_server(self, dataset_id: int = None, limit: int = 50):
        items = []
        if dataset_id is not None:
            infos = self._api.image.get_list(dataset_id, limit=limit)
            items = [Select.Item(value=info.id, label=info.name) for info in infos]
        return items

    def refresh_items(self, dataset_id: int = None, limit: int = 50):
        items = self._get_items_from_server(dataset_id, limit)
        self._item_selector.set(items=items)
        need_limit, ds_items_count = self._need_set_limit(dataset_id)
        if need_limit is True:
            self._set_limit(ds_items_count)



# coding: utf-8

# docs
import os
from collections import namedtuple
from supervisely.api.api import Api
from supervisely.annotation.annotation import Annotation
from typing import Tuple, List, Dict, Optional, Callable

from supervisely._utils import batched
from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.io.fs import touch, dir_exists, list_files, mkdir
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import PointcloudEpisodeAnnotation
from supervisely.project.pointcloud_project import PointcloudProject, PointcloudDataset
from supervisely.project.project import OpenMode
from supervisely.project.project import read_single_project as read_project_wrapper
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import Progress
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.project.project_type import ProjectType

PointcloudItemPaths = namedtuple('PointcloudItemPaths', ['pointcloud_path', 'related_images_dir'])


class PointcloudEpisodeDataset(PointcloudDataset):
    item_dir_name = 'pointcloud'
    related_images_dir_name = 'related_images'
    annotation_class = PointcloudEpisodeAnnotation

    def get_item_paths(self, item_name: str) -> PointcloudItemPaths:
        return PointcloudItemPaths(pointcloud_path=self.get_img_path(item_name),
                                   related_images_dir=self.get_related_images_path(item_name))

    def get_ann_path(self) -> str:
        return os.path.join(self.directory, "annotation.json")

    def get_frame_pointcloud_map_path(self) -> str:
        return os.path.join(self.directory, "frame_pointcloud_map.json")

    def set_ann(self, ann: Annotation) -> None:
        if type(ann) is not self.annotation_class:
            raise TypeError("Type of 'ann' have to be Annotation, not a {}".format(type(ann)))
        dst_ann_path = self.get_ann_path()
        dump_json_file(ann.to_json(), dst_ann_path)

    def _add_ann_by_type(self, item_name, ann):
        return

    def _add_img_info(self, item_name, img_info=None):
        return

    def _create(self):
        mkdir(self.item_dir)

    def _read(self):
        if not dir_exists(self.item_dir):
            raise NotADirectoryError(f"Cannot read dataset {self.item_dir}: directory not found")

        try:
            item_paths = sorted(list_files(self.item_dir, filter_fn=self._has_valid_ext))
            item_names = sorted([os.path.basename(path) for path in item_paths])

            map_file_path = self.get_frame_pointcloud_map_path()
            if os.path.isfile(map_file_path):
                self._frame_to_pc_map = load_json_file(map_file_path)
            else:
                self._frame_to_pc_map = {frame_index: item_names[frame_index] for frame_index in range(len(item_names))}

            self._pc_to_frame = {v: k for k, v in self._frame_to_pc_map.items()}
            self._item_to_ann = {name: self._pc_to_frame[name] for name in item_names}
        except Exception as ex:
            raise Exception(f"Cannot read dataset ({self.item_dir}): {repr(ex)}")

    def get_frame_idx(self, item_name: str) -> int:
        return int(self._item_to_ann[item_name])


class PointcloudEpisodeProject(PointcloudProject):
    dataset_class = PointcloudEpisodeDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = PointcloudEpisodeDataset

    @classmethod
    def read_single(cls, dir):
        return read_project_wrapper(dir, cls)


def download_pointcloud_episode_project(api: Api, project_id: int, dest_dir: str, dataset_ids: Optional[List[int]]=None,
                                        download_pcd: Optional[bool]=True,
                                        download_related_images: Optional[bool]=True,
                                        download_annotations: Optional[bool]=True,
                                        log_progress: Optional[bool]=False, batch_size: Optional[int]=10,
                                        progress_cb: Optional[Callable] = None) -> None:
    key_id_map = KeyIdMap()
    project_fs = PointcloudEpisodeProject(dest_dir, OpenMode.CREATE)
    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs = project_fs.create_dataset(dataset.name)
        pointclouds = api.pointcloud_episode.get_list(dataset.id)

        if download_annotations:
            # Download annotation to project_path/dataset_path/annotation.json
            ann_json = api.pointcloud_episode.annotation.download(dataset.id)
            annotation = dataset_fs.annotation_class.from_json(ann_json, meta, key_id_map)
            dataset_fs.set_ann(annotation)

            # frames --> pointcloud mapping to project_path/dataset_path/frame_pointcloud_map.json
            frame_name_map = api.pointcloud_episode.get_frame_name_map(dataset.id)
            frame_pointcloud_map_path = dataset_fs.get_frame_pointcloud_map_path()
            dump_json_file(frame_name_map, frame_pointcloud_map_path)

        # Download data
        if log_progress:
            ds_progress = Progress('Downloading dataset: {!r}'.format(dataset.name), total_cnt=len(pointclouds))

        for batch in batched(pointclouds, batch_size=batch_size):
            pointcloud_ids = [pointcloud_info.id for pointcloud_info in batch]
            pointcloud_names = [pointcloud_info.name for pointcloud_info in batch]

            for pointcloud_id, pointcloud_name in zip(pointcloud_ids, pointcloud_names):
                pointcloud_file_path = dataset_fs.generate_item_path(pointcloud_name)
                if download_pcd is True:
                    api.pointcloud_episode.download_path(pointcloud_id, pointcloud_file_path)
                else:
                    touch(pointcloud_file_path)

                if download_related_images:
                    related_images_path = dataset_fs.get_related_images_path(pointcloud_name)
                    related_images = api.pointcloud_episode.get_list_related_images(pointcloud_id)
                    for rimage_info in related_images:
                        name = rimage_info[ApiField.NAME]
                        rimage_id = rimage_info[ApiField.ID]

                        path_img = os.path.join(related_images_path, name)
                        path_json = os.path.join(related_images_path, name + ".json")

                        api.pointcloud_episode.download_related_image(rimage_id, path_img)
                        dump_json_file(rimage_info, path_json)

                dataset_fs.add_item_file(pointcloud_name,
                                         pointcloud_file_path,
                                         _validate_item=False)
            if progress_cb is not None:
                progress_cb(len(batch))
            if log_progress:
                ds_progress.iters_done_report(len(batch))

    project_fs.set_key_id_map(key_id_map)


def upload_pointcloud_episode_project(directory: str, api: Api, workspace_id: int, project_name: Optional[str]=None,
                                      log_progress: Optional[bool]=False) -> Tuple[int, str]:
    # STEP 0 — create project remotely
    project_locally = PointcloudEpisodeProject(directory, OpenMode.READ)
    project_name = project_locally.name if project_name is None else project_name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project_remotely = api.project.create(workspace_id, project_name, ProjectType.POINT_CLOUD_EPISODES)
    api.project.update_meta(project_remotely.id, project_locally.meta.to_json())

    uploaded_objects = KeyIdMap()
    for dataset_locally in project_locally.datasets:
        ann_json_path = dataset_locally.get_ann_path()

        if os.path.isfile(ann_json_path):
            ann_json = load_json_file(ann_json_path)
            episode_annotation = PointcloudEpisodeAnnotation.from_json(ann_json, project_locally.meta)
        else:
            episode_annotation = PointcloudEpisodeAnnotation()

        dataset_remotely = api.dataset.create(project_remotely.id,
                                              dataset_locally.name,
                                              description=episode_annotation.description,
                                              change_name_if_conflict=True)

        # STEP 1 — upload episodes
        items_infos = {
            'names': [],
            'paths': [],
            'metas': []
        }

        for item_name in dataset_locally:
            item_path, related_images_dir = dataset_locally.get_item_paths(item_name)
            frame_idx = dataset_locally.get_frame_idx(item_name)

            item_meta = {
                "frame": frame_idx
            }

            items_infos['names'].append(item_name)
            items_infos['paths'].append(item_path)
            items_infos['metas'].append(item_meta)

        ds_progress = Progress('Uploading pointclouds: {!r}'.format(dataset_remotely.name),
                               total_cnt=len(dataset_locally)) if log_progress else None
        pcl_infos = api.pointcloud_episode.upload_paths(dataset_remotely.id,
                                                        names=items_infos['names'],
                                                        paths=items_infos['paths'],
                                                        metas=items_infos['metas'],
                                                        progress_cb=ds_progress.iters_done_report if log_progress else None)

        # STEP 2 — upload annotations
        frame_to_pcl_ids = {pcl_info.frame: pcl_info.id for pcl_info in pcl_infos}
        api.pointcloud_episode.annotation.append(dataset_remotely.id,
                                                 episode_annotation,
                                                 frame_to_pcl_ids,
                                                 uploaded_objects)

        # STEP 3 — upload photo context
        img_infos = {
            'img_paths': [],
            'img_metas': []
        }

        # STEP 3.1 — upload images
        for pcl_info in pcl_infos:
            related_items = dataset_locally.get_related_images(pcl_info.name)
            images_paths_for_frame = [img_path for img_path, _ in related_items]

            img_infos['img_paths'].extend(images_paths_for_frame)

        img_progress = Progress('Uploading photo context: {!r}'.format(dataset_remotely.name),
                                total_cnt=len(img_infos['img_paths'])) if log_progress else None

        images_hashes = api.pointcloud_episode.upload_related_images(img_infos['img_paths'],
                                                                     progress_cb=img_progress.iters_done_report if log_progress else None)

        # STEP 3.2 — upload images metas
        images_hashes_iterator = images_hashes.__iter__()
        for pcl_info in pcl_infos:
            related_items = dataset_locally.get_related_images(pcl_info.name)

            for _, meta_json in related_items:
                img_hash = next(images_hashes_iterator)
                img_infos['img_metas'].append({ApiField.ENTITY_ID: pcl_info.id,
                                               ApiField.NAME: meta_json[ApiField.NAME],
                                               ApiField.HASH: img_hash,
                                               ApiField.META: meta_json[ApiField.META]})

        if len(img_infos['img_metas']) > 0:
            api.pointcloud_episode.add_related_images(img_infos['img_metas'])

    return project_remotely.id, project_remotely.name



import os
import numpy
import functools
from supervisely.sly_logger import logger
from supervisely.io.fs import silent_remove
from supervisely.geometry.bitmap import Bitmap
from supervisely.imaging import image as sly_image
from supervisely.geometry.rectangle import Rectangle
from supervisely._utils import rand_str as sly_rand_str
from supervisely.annotation.annotation import Annotation
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.sliding_windows_fuzzy import (
    SlidingWindowsFuzzy,
    SlidingWindowBorderStrategy,
)


def process_image_roi(func):
    """
    Decorator for processing annotation labels before and after inference.

    Crops input image before inference if kwargs['state']['rectangle_crop'] provided
    and then scales annotation back to original image size.

    Keyword arguments:

    :param image_np: Image in numpy.ndarray format (use image_path or image_np, not both)
    :type image_np: numpy.ndarray
    :param image_path: Path to image (use image_path or image_np, not both)
    :type image_path: str
    :param project_meta: ProjectMeta of the current project
    :type project_meta: ProjectMeta
    :param state: Application state
    :type state: dict
    :raises: :class:`ValueError`, if image_np or image_path invalid or not provided
    :return: Annotation in json format
    :rtype: :class:`dict`
    """

    def process_image_path(image_path, sly_rectangle):
        image = sly_image.read(image_path)
        image_size = image.shape[:2]
        image_base_dir = os.path.dirname(image_path)
        image_name, image_ext = os.path.splitext(os.path.basename(image_path))

        image_crop = sly_image.crop(image, sly_rectangle)
        image_crop_path = os.path.join(
            image_base_dir, sly_rand_str(10) + "_" + image_name + "_crop" + image_ext
        )
        sly_image.write(image_crop_path, image_crop)
        return image_crop_path, image_size

    def process_image_np(image_np, sly_rectangle):
        image_crop = sly_image.crop(image_np, sly_rectangle)
        image_size, image_crop_size = image_np.shape[:2], image_crop.shape[:2]
        return image_crop, image_size

    def scale_ann_to_original_size(
        ann_json, project_meta, original_size, sly_rectangle
    ):
        ann = Annotation.from_json(ann_json, project_meta)
        updated_labels = []
        for label in ann.labels:
            if type(label.geometry) is Rectangle:
                updated_geometry = Rectangle(
                    top=label.geometry.top + sly_rectangle.top,
                    left=label.geometry.left + sly_rectangle.left,
                    bottom=label.geometry.bottom + sly_rectangle.top,
                    right=label.geometry.right + sly_rectangle.left,
                )

            if type(label.geometry) is Bitmap:
                bitmap_data = label.geometry.data
                bitmap_origin = PointLocation(
                    label.geometry.origin.row + sly_rectangle.top,
                    label.geometry.origin.col + sly_rectangle.left,
                )

                updated_geometry = Bitmap(data=bitmap_data, origin=bitmap_origin)
            updated_labels.append(label.clone(geometry=updated_geometry))

        ann = ann.clone(img_size=original_size, labels=updated_labels)
        ann_json = ann.to_json()
        return ann_json

    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        project_meta = kwargs["project_meta"]
        state = kwargs["state"]
        rectangle_json = state.get("rectangle")

        if rectangle_json is None:
            ann_json = func(*args, **kwargs)
            return ann_json

        rectangle = Rectangle.from_json(rectangle_json)
        if "image_np" in kwargs.keys():
            image_np = kwargs["image_np"]
            if not isinstance(image_np, numpy.ndarray):
                raise ValueError("Invalid input. Image path must be numpy.ndarray")
            image_crop_np, image_size = process_image_np(image_np, rectangle)
            kwargs["image_np"] = image_crop_np
            ann_json = func(*args, **kwargs)
            ann_json = scale_ann_to_original_size(
                ann_json, project_meta, image_size, rectangle
            )
        elif "image_path" in kwargs.keys():
            image_path = kwargs["image_path"]
            if not isinstance(image_path, str):
                raise ValueError("Invalid input. Image path must be str")
            image_crop_path, image_size = process_image_path(image_path, rectangle)
            kwargs["image_path"] = image_crop_path
            ann_json = func(*args, **kwargs)
            ann_json = scale_ann_to_original_size(
                ann_json, project_meta, image_size, rectangle
            )
            silent_remove(image_crop_path)
        else:
            raise ValueError("image_np or image_path not provided!")

        return ann_json

    return wrapper_inference


def process_image_sliding_window(func):
    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        project_meta = kwargs["project_meta"]
        state = kwargs["state"]
        settings = kwargs["settings"]
        inference_mode = settings.get("inference_mode", "full_image")
        if inference_mode != "sliding_window":
            ann_json = func(*args, **kwargs)
            return ann_json

        sliding_window_params = settings["sliding_window_params"]
        image_path = kwargs["image_path"]
        img = sly_image.read(image_path)
        img_h, img_w = img.shape[:2]
        windowHeight = sliding_window_params.get("windowHeight", img_h)
        windowWidth = sliding_window_params.get("windowWidth", img_w)
        overlapY = sliding_window_params.get("overlapY", 0)
        overlapX = sliding_window_params.get("overlapX", 0)
        borderStrategy = sliding_window_params.get("borderStrategy", "shift_window")

        slider = SlidingWindowsFuzzy(
            [windowHeight, windowWidth], [overlapY, overlapX], borderStrategy
        )
        rectangles = []
        for window in slider.get(img.shape[:2]):
            rectangles.append(window)

        ann = Annotation.from_img_path(image_path)
        results = []
        for rect in rectangles:
            state["rectangle"] = rect.to_json()
            ann_json = func(*args, **kwargs)
            slice_ann = Annotation.from_json(ann_json, project_meta)
            ann = ann.add_labels(slice_ann.labels)
            results.append(
                {
                    "rectangle": rect.to_json(),
                    "labels": [l.to_json() for l in slice_ann.labels],
                }
            )
        return {"annotation": ann.to_json(), "data": {"slides": results}}

    return wrapper_inference



import os
import signal
import psutil
import sys
from pathlib import Path

from fastapi import (
    FastAPI,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    Depends,
    HTTPException,
)
import jinja2
from fastapi.testclient import TestClient
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from supervisely.app.fastapi.utils import run_sync

from supervisely.app.singleton import Singleton

from supervisely.app.fastapi.templating import Jinja2Templates
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.io.fs import mkdir, dir_exists
from supervisely.sly_logger import logger
from supervisely.api.api import SERVER_ADDRESS, API_TOKEN, TASK_ID, Api
from supervisely._utils import is_production, is_development
from async_asgi_testclient import TestClient


def create(process_id=None, headless=False) -> FastAPI:
    from supervisely.app import DataJson, StateJson

    app = FastAPI()
    WebsocketManager().set_app(app)

    @app.post("/shutdown")
    async def shutdown_endpoint(request: Request):
        shutdown(process_id)

    if headless is False:

        @app.post("/data")
        async def send_data(request: Request):
            data = DataJson()
            response = JSONResponse(content=dict(data))
            return response

        @app.post("/state")
        async def send_state(request: Request):
            state = StateJson()
            response = JSONResponse(content=dict(state))

            gettrace = getattr(sys, "gettrace", None)
            if (gettrace is not None and gettrace()) or is_development():
                response.headers["x-debug-mode"] = "1"
            return response

        @app.post("/session-info")
        async def send_session_info(request: Request):
            server_address = os.environ.get(SERVER_ADDRESS)
            if server_address is not None:
                server_address = Api.normalize_server_address(server_address)

            response = JSONResponse(
                content={
                    TASK_ID: os.environ.get(TASK_ID),
                    SERVER_ADDRESS: server_address,
                    API_TOKEN: os.environ.get(API_TOKEN),
                }
            )
            return response

        @app.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            await WebsocketManager().connect(websocket)
            try:
                while True:
                    data = await websocket.receive_text()
            except WebSocketDisconnect:
                WebsocketManager().disconnect(websocket)

        import supervisely

        app.mount(
            "/css", StaticFiles(directory=supervisely.__path__[0]), name="sly_static"
        )

    return app


def shutdown(process_id=None):
    try:
        logger.info(f"Shutting down [pid argument = {process_id}]...")
        if process_id is None:
            # process_id = psutil.Process(os.getpid()).ppid()
            process_id = os.getpid()
        current_process = psutil.Process(process_id)
        current_process.send_signal(signal.SIGINT)  # emit ctrl + c
    except KeyboardInterrupt:
        logger.info("Application has been shut down successfully")


def enable_hot_reload_on_debug(app: FastAPI):
    templates = Jinja2Templates()
    gettrace = getattr(sys, "gettrace", None)
    if gettrace is None:
        print("Can not detect debug mode, no sys.gettrace")
    elif gettrace():
        import arel

        hot_reload = arel.HotReload(paths=[arel.Path(".")])
        app.add_websocket_route("/hot-reload", route=hot_reload, name="hot-reload")
        app.add_event_handler("startup", hot_reload.startup)
        app.add_event_handler("shutdown", hot_reload.shutdown)
        templates.env.globals["DEBUG"] = "1"
        templates.env.globals["hot_reload"] = hot_reload
        logger.debug("Debugger (gettrace) detected, UI hot-reload is enabled")
    else:
        logger.debug("In runtime mode ...")


def handle_server_errors(app: FastAPI):
    @app.exception_handler(500)
    async def server_exception_handler(request, exc):
        return await http_exception_handler(
            request,
            HTTPException(
                status_code=500,
                detail={
                    # "title": "error title",
                    "message": repr(exc)
                },
            ),
        )


def _init(
    app: FastAPI = None,
    templates_dir: str = "templates",
    headless=False,
    process_id=None,
) -> FastAPI:
    from supervisely.app.fastapi import available_after_shutdown
    from supervisely.app.content import StateJson

    if app is None:
        app = _MainServer().get_server()

    handle_server_errors(app)

    if headless is False:
        if "app_body_padding" not in StateJson():
            StateJson()["app_body_padding"] = "20px"
        Jinja2Templates(
            directory=[str(Path(__file__).parent.absolute()), templates_dir]
        )
        enable_hot_reload_on_debug(app)

    app.mount("/sly", create(process_id, headless))

    if headless is False:

        @app.middleware("http")
        async def get_state_from_request(request: Request, call_next):
            await StateJson.from_request(request)
            response = await call_next(request)
            return response

        @app.get("/")
        @available_after_shutdown(app)
        def read_index(request: Request):
            return Jinja2Templates().TemplateResponse(
                "index.html", {"request": request}
            )

        @app.on_event("shutdown")
        def shutdown():
            client = TestClient(app)
            resp = run_sync(client.get("/"))
            assert resp.status_code == 200
            logger.info("Application has been shut down successfully")

    return app


class _MainServer(metaclass=Singleton):
    def __init__(self):
        self._server = FastAPI()

    def get_server(self) -> FastAPI:
        return self._server


class Application(metaclass=Singleton):
    def __init__(self, templates_dir: str = "templates", headless=False):
        if is_production():
            logger.info(
                "Application is running on Supervisely Platform in production mode"
            )
        else:
            logger.info("Application is running on localhost in development mode")
        self._process_id = os.getpid()
        logger.info(f"Application PID is {self._process_id}")
        self._fastapi: FastAPI = _init(
            app=None,
            templates_dir=templates_dir,
            headless=headless,
            process_id=self._process_id,
        )

    def get_server(self):
        return self._fastapi

    async def __call__(self, scope, receive, send) -> None:
        await self._fastapi.__call__(scope, receive, send)

    def shutdown(self):
        shutdown(self._process_id)


def get_name_from_env(default="Supervisely App"):
    return os.environ.get("APP_NAME", default)



from pathlib import Path
from bs4 import BeautifulSoup
import re
import time
import uuid
from typing import Union
from varname import varname
from jinja2 import Environment
import markupsafe
from supervisely.app.jinja2 import create_env
from supervisely.app.content import DataJson, StateJson
from fastapi import FastAPI
from supervisely.app.fastapi import _MainServer
from supervisely.app.widgets_context import JinjaWidgets
from supervisely._utils import generate_free_name, rand_str


class Hidable:
    def __init__(self):
        self._hide = False

    def hide(self):
        self._hide = True
        DataJson()[self.widget_id]["hide"] = self._hide
        DataJson().send_changes()

    def show(self):
        self._hide = False
        DataJson()[self.widget_id]["hide"] = self._hide
        DataJson().send_changes()

    def get_json_data(self):
        return {"hide": self._hide}

    def get_json_state(self):
        raise {}

    def _wrap_hide_html(self, widget_id, html):
        return f'<div v-if="!data.{widget_id}.hide">{html}</div>'


class Disableable:
    def __init__(self):
        self._disabled = False

    def disable(self):
        self._disabled = True
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._disabled = False
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def get_json_data(self):
        return {"disabled": self._disabled}

    def get_json_state(self):
        raise {}

    def _wrap_disable_html(self, widget_id, html):
        soup = BeautifulSoup(html, features="html.parser")
        results = soup.find_all(re.compile("^el-"))
        for tag in results:
            if not tag.has_attr("disabled") and not tag.has_attr(":disabled"):
                tag[":disabled"] = f"data.{widget_id}.disabled"
        return str(soup)


def generate_id():
    return "auto" + uuid.uuid4().hex


class Widget(Hidable, Disableable):
    def __init__(self, widget_id: str = None, file_path: str = __file__):
        super().__init__()
        self._sly_app = _MainServer()
        self.widget_id = widget_id
        self._file_path = file_path
        if self.widget_id is None:
            try:
                self.widget_id = varname(frame=2)
            except Exception as e:  # Caller doesn\\\'t assign the result directly to variable(s).
                try:
                    self.widget_id = varname(frame=3)
                except Exception as e:  # VarnameRetrievingError('Unable to retrieve the ast node.')
                    self.widget_id = type(self).__name__ + rand_str(10)

        self._register()

    def _register(self):
        # get singletons
        data = DataJson()
        data.raise_for_key(self.widget_id)
        self.update_data()

        state = StateJson()
        state.raise_for_key(self.widget_id)
        self.update_state(state=state)

        JinjaWidgets().context[self.widget_id] = self
        # templates = Jinja2Templates()
        # templates.context_widgets[self.widget_id] = self

    def get_json_data(self):
        raise NotImplementedError()

    def get_json_state(self):
        raise NotImplementedError()

    def update_state(self, state=None):
        serialized_state = self.get_json_state()
        if serialized_state is not None:
            if state is None:
                state = StateJson()
            state.setdefault(self.widget_id, {}).update(serialized_state)

    def update_data(self):
        data = DataJson()

        widget_data = self.get_json_data()
        if widget_data is None:
            widget_data = {}
        hidable_data = super().get_json_data()
        disableable_data = super().get_json_data()

        serialized_data = {**widget_data, **hidable_data, **disableable_data}
        if serialized_data is not None:
            data.setdefault(self.widget_id, {}).update(serialized_data)

    def get_route_path(self, route: str) -> str:
        return f"/{self.widget_id}/{route}"

    def add_route(self, app, route):
        def decorator(f):
            existing_cb = DataJson()[self.widget_id].get("widget_routes", {}).get(route)
            if existing_cb is not None:
                raise Exception(
                    f"Route [{route}] already attached to function with name: {existing_cb}"
                )

            app.add_api_route(f"/{self.widget_id}/{route}", f, methods=["POST"])
            DataJson()[self.widget_id].setdefault("widget_routes", {})[
                route
            ] = f.__name__

            self.update_data()

        return decorator

    def to_html(self):
        current_dir = Path(self._file_path).parent.absolute()
        jinja2_sly_env: Environment = create_env(current_dir)
        html = jinja2_sly_env.get_template("template.html").render({"widget": self})
        # st = time.time()
        res = self._wrap_disable_html(self.widget_id, html)
        # print("---> Time (_wrap_disable_html): ", time.time() - st, " seconds")
        # st = time.time()
        res = self._wrap_hide_html(self.widget_id, res)
        # print("---> time (_wrap_hide_html): ", time.time() - st, " seconds")
        return markupsafe.Markup(res)

    def __html__(self):
        res = self.to_html()
        return res


# https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string
# https://github.com/pwwang/python-varname
# https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string/18425523#18425523
# https://ideone.com/ym3bkD
# https://github.com/pwwang/python-varname
# https://stackoverflow.com/questions/13034496/using-global-variables-between-files
# https://docs.python.org/3/faq/programming.html#how-do-i-share-global-variables-across-modules



# coding: utf-8

# docs
from requests import Response
from typing import List, NamedTuple, Dict, Optional, Callable
from supervisely.task.progress import Progress

from collections import defaultdict
from supervisely.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely.io.fs import ensure_base_path, get_file_hash
from supervisely._utils import batched

from supervisely.api.pointcloud.pointcloud_annotation_api import PointcloudAnnotationAPI
from supervisely.api.pointcloud.pointcloud_object_api import PointcloudObjectApi
from supervisely.api.pointcloud.pointcloud_figure_api import PointcloudFigureApi
from supervisely.api.pointcloud.pointcloud_tag_api import PointcloudTagApi
from requests_toolbelt import MultipartDecoder, MultipartEncoder


class PointcloudInfo(NamedTuple):
    id: int
    frame: int
    description: str
    name: str
    team_id: int
    workspace_id: int
    project_id: int
    dataset_id: int
    link: str
    hash: str
    path_original: str
    cloud_mime: str
    figures_count: int
    objects_count: int
    tags: list
    meta: dict
    created_at: str
    updated_at: str


class PointcloudApi(RemoveableBulkModuleApi):
    def __init__(self, api):
        """
        :param api: Api class object
        """
        super().__init__(api)
        self.annotation = PointcloudAnnotationAPI(api)
        self.object = PointcloudObjectApi(api)
        self.figure = PointcloudFigureApi(api)
        self.tag = PointcloudTagApi(api)

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.FRAME,
                ApiField.DESCRIPTION,
                ApiField.NAME,
                ApiField.TEAM_ID,
                ApiField.WORKSPACE_ID,
                ApiField.PROJECT_ID,
                ApiField.DATASET_ID,
                ApiField.LINK,
                ApiField.HASH,
                ApiField.PATH_ORIGINAL,
                # ApiField.PREVIEW,
                ApiField.CLOUD_MIME,
                ApiField.FIGURES_COUNT,
                ApiField.ANN_OBJECTS_COUNT,
                ApiField.TAGS,
                ApiField.META,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'PointCloudInfo'

    def _convert_json_info(self, info: Dict, skip_missing: Optional[bool] = True):
        res = super(PointcloudApi, self)._convert_json_info(info, skip_missing=skip_missing)
        return PointcloudInfo(**res._asdict())

    def get_list(self, dataset_id: int, filters: Optional[List[Dict[str, str]]] = None) -> List[PointcloudInfo]:
        """
        :param dataset_id: int
        :param filters: list
        :return: list of the pointclouds objects from the dataset with given id
        """
        return self.get_list_all_pages('point-clouds.list',
                                       {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id: int) -> PointcloudInfo:
        """
        :param id: int
        :return: PointcloudApi metadata by numeric id
        """
        return self._get_info_by_id(id, 'point-clouds.info')

    def _download(self, id: int, is_stream: Optional[bool] = False):
        """
        :param id: int
        :param is_stream: bool
        :return: Response object containing pointcloud object with given id
        """
        response = self._api.post('point-clouds.download', {ApiField.ID: id}, stream=is_stream)
        return response

    def download_path(self, id: int, path: str) -> None:
        """
        Download pointcloud with given id on the given path
        :param id: int
        :param path: str
        """
        response = self._download(id, is_stream=True)
        ensure_base_path(path)
        with open(path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)

    def get_list_related_images(self, id: int) -> List:
        dataset_id = self.get_info_by_id(id).dataset_id
        filters = [{"field": ApiField.ENTITY_ID, "operator": "=", "value": id}]
        return self.get_list_all_pages('point-clouds.images.list',
                                       {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters},
                                       convert_json_info_cb=lambda x: x)

    def download_related_image(self, id: int, path: str) -> Response:
        response = self._api.post('point-clouds.images.download', {ApiField.ID: id}, stream=True)
        ensure_base_path(path)
        with open(path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
        return response

    # @TODO: copypaste from video_api
    def upload_hash(self, dataset_id: int, name: str, hash: str, meta: Optional[Dict] = None) -> PointcloudInfo:
        meta = {} if meta is None else meta
        return self.upload_hashes(dataset_id, [name], [hash], [meta])[0]

    # @TODO: copypaste from video_api
    def upload_hashes(self, dataset_id: int, names: List[str], hashes: List[str], metas: Optional[List[Dict]] = None,
                      progress_cb: Optional[Callable] = None) -> List[PointcloudInfo]:
        return self._upload_bulk_add(lambda item: (ApiField.HASH, item), dataset_id, names, hashes, metas, progress_cb)

    # @TODO: copypaste from video_api
    def _upload_bulk_add(self, func_item_to_kv, dataset_id, names, items, metas=None, progress_cb=None):
        if metas is None:
            metas = [{}] * len(items)

        results = []
        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError("Can not match \"names\" and \"items\" lists, len(names) != len(items)")

        for batch in batched(list(zip(names, items, metas))):
            images = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                images.append({ApiField.NAME: name,
                               item_tuple[0]: item_tuple[1],
                               ApiField.META: meta if meta is not None else {}})
            response = self._api.post('point-clouds.bulk.add', {ApiField.DATASET_ID: dataset_id,
                                                                ApiField.POINTCLOUDS: images})
            if progress_cb is not None:
                progress_cb(len(images))

            results.extend([self._convert_json_info(item) for item in response.json()])
        name_to_res = {img_info.name: img_info for img_info in results}
        ordered_results = [name_to_res[name] for name in names]

        return ordered_results

    def upload_related_image(self, path: str) -> List[str]:
        return self.upload_related_images([path])

    def upload_related_images(self, paths: List[str], progress_cb: Optional[Callable] = None) -> List[str]:
        def path_to_bytes_stream(path):
            return open(path, 'rb')

        return self._upload_data_bulk(path_to_bytes_stream, get_file_hash, paths, progress_cb)

    def add_related_images(self, images_json: Dict) -> Dict:
        response = self._api.post('point-clouds.images.add', {ApiField.IMAGES: images_json})
        return response.json()

    def upload_path(self, dataset_id: int, name: str, path: str, meta: Optional[Dict] = None) -> PointcloudInfo:
        metas = None if meta is None else [meta]
        return self.upload_paths(dataset_id, [name], [path], metas=metas)[0]

    def upload_paths(self, dataset_id: int, names: List[str], paths: List[str], progress_cb: Optional[Callable] = None,
                     metas: Optional[Dict] = None) -> List[PointcloudInfo]:
        def path_to_bytes_stream(path):
            return open(path, 'rb')

        hashes = self._upload_data_bulk(path_to_bytes_stream, get_file_hash, paths, progress_cb)
        return self.upload_hashes(dataset_id, names, hashes, metas=metas)

    def check_existing_hashes(self, hashes: List[str]) -> List[str]:
        results = []
        if len(hashes) == 0:
            return results
        for hashes_batch in batched(hashes, batch_size=900):
            response = self._api.post('images.internal.hashes.list', hashes_batch)
            results.extend(response.json())
        return results

    def _upload_data_bulk(self, func_item_to_byte_stream, func_item_hash, items, progress_cb):
        hashes = []
        if len(items) == 0:
            return hashes

        hash_to_items = defaultdict(list)

        for idx, item in enumerate(items):
            item_hash = func_item_hash(item)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(item)

        unique_hashes = set(hashes)
        remote_hashes = self.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        # upload only new images to supervisely server
        items_to_upload = []
        for hash in new_hashes:
            items_to_upload.extend(hash_to_items[hash])

        for batch in batched(items_to_upload):
            content_dict = {}
            for idx, item in enumerate(batch):
                content_dict["{}-file".format(idx)] = (str(idx), func_item_to_byte_stream(item), 'pcd/*')
            encoder = MultipartEncoder(fields=content_dict)
            self._api.post('point-clouds.bulk.upload', encoder)
            if progress_cb is not None:
                progress_cb(len(batch))

        return hashes



from supervisely.app.widgets import Widget
from typing import Dict, List


class Flexbox(Widget):
    # https://www.w3schools.com/css/css3_flexbox.asp
    def __init__(
        self,
        widgets: List[Widget],  # or RadioGroup in future
        widget_id: str = None,
    ):
        self._widgets = widgets
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {}



import copy

import numpy as np
import pandas as pd
from typing import NamedTuple, Any, List

from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger


class PackerUnpacker:
    SUPPORTED_TYPES = tuple([dict, pd.DataFrame])

    @staticmethod
    def validate_sizes(unpacked_data):
        for row in unpacked_data["data"]:
            if len(row) != len(unpacked_data["columns"]):
                raise ValueError(
                    "Sizes mismatch:\n"
                    f'{len(row)} != {len(unpacked_data["columns"])}\n'
                    f"{row}\n"
                    f'{unpacked_data["columns"]}'
                )

    @staticmethod
    def unpack_data(data, unpacker_cb):
        unpacked_data = unpacker_cb(data)
        PackerUnpacker.validate_sizes(unpacked_data)
        return unpacked_data

    @staticmethod
    def pack_data(data, packer_cb):
        packed_data = packer_cb(data)
        return packed_data

    @staticmethod
    def dict_unpacker(data: dict):
        unpacked_data = {"columns": data["columns"], "data": data["data"]}

        return unpacked_data

    @staticmethod
    def pandas_unpacker(data: pd.DataFrame):
        data = data.where(pd.notnull(data), None)
        data = data.astype(object).replace(np.nan, "-")  # may be None

        unpacked_data = {
            "columns": data.columns.to_list(),
            "data": data.values.tolist(),
        }
        return unpacked_data

    @staticmethod
    def dict_packer(data):
        packed_data = {"columns": data["columns"], "data": data["data"]}
        return packed_data

    @staticmethod
    def pandas_packer(data):
        packed_data = pd.DataFrame(data=data["data"], columns=data["columns"])
        return packed_data


DATATYPE_TO_PACKER = {
    pd.DataFrame: PackerUnpacker.pandas_packer,
    dict: PackerUnpacker.dict_packer,
}

DATATYPE_TO_UNPACKER = {
    pd.DataFrame: PackerUnpacker.pandas_unpacker,
    dict: PackerUnpacker.dict_unpacker,
}

"""
iris = pd.read_csv(
    "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv"
)
iris.insert(loc=0, column="index", value=np.arange(len(iris)))
table = sly.app.widgets.Table(data=iris, fixed_cols=1)

@table.click
def show_image(datapoint: sly.app.widgets.Table.ClickedDataPoint):
    print("Column name = ", datapoint.column_name)
    print("Cell value = ", datapoint.cell_value)
    print("Row = ", datapoint.row)

"""


class Table(Widget):
    class Routes:
        CELL_CLICKED = "cell_clicked_cb"

    class ClickedDataPoint(NamedTuple):
        column_index: int
        column_name: str
        cell_value: Any
        row: dict

    def __init__(
        self,
        data=None,
        columns: list = None,
        fixed_cols: int = None,
        per_page: int = 10,
        page_sizes: List[int] = [10, 15, 30, 50, 100],
        width: str = "auto",  # "200px", or "100%"
        widget_id: str = None,
    ):
        """
        :param data: Data of table in different formats:
        1. Pandas Dataframe or pd.DataFrame(data=data, columns=columns)
        2. Python dict with structure {
                                        'columns_names': ['col_name_1', 'col_name_2', ...],
                                        'values_by_rows': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
        """

        self._supported_types = PackerUnpacker.SUPPORTED_TYPES

        self._parsed_data = None
        self._data_type = None
        self._click_handled = False

        self._update_table_data(input_data=pd.DataFrame(data=data, columns=columns))

        self._per_page = per_page
        self._page_sizes = page_sizes
        self._fix_columns = fixed_cols
        self._width = width
        self._loading = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "table_data": self._parsed_data,
            "table_options": {
                "perPage": self._per_page,
                "pageSizes": self._page_sizes,
                "fixColumns": self._fix_columns,
            },
            "loading": self._loading,
        }

    def get_json_state(self):
        return {"selected_row": {}}

    def _update_table_data(self, input_data):
        if input_data is not None:
            self._parsed_data = copy.deepcopy(
                self._get_unpacked_data(input_data=input_data)
            )
        else:
            self._parsed_data = {"columns": [], "data": []}
            self._data_type = dict

    def _get_packed_data(self, input_data, data_type):
        return PackerUnpacker.pack_data(
            data=input_data, packer_cb=DATATYPE_TO_PACKER[data_type]
        )

    def _get_unpacked_data(self, input_data):
        input_data_type = type(input_data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
                            """
            )

        return PackerUnpacker.unpack_data(
            data=input_data, unpacker_cb=DATATYPE_TO_UNPACKER[input_data_type]
        )

    @property
    def fixed_columns_num(self):
        return self._fix_columns

    @fixed_columns_num.setter
    def fixed_columns_num(self, value):
        self._fix_columns = value
        DataJson()[self.widget_id]["table_options"]["fixColumns"] = self._fix_columns

    def to_json(self) -> dict:
        return self._get_packed_data(self._parsed_data, dict)

    def to_pandas(self) -> pd.DataFrame:
        return self._get_packed_data(self._parsed_data, pd.DataFrame)

    def read_json(self, value: dict):
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data
        DataJson().send_changes()
        self.clear_selection()

    def read_pandas(self, value: pd.DataFrame):
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data
        DataJson().send_changes()
        self.clear_selection()

    def insert_row(self, data, index=-1):
        PackerUnpacker.validate_sizes(
            {"columns": self._parsed_data["columns"], "data": [data]}
        )

        table_data = self._parsed_data["data"]
        index = len(table_data) if index > len(table_data) or index < 0 else index

        self._parsed_data["data"].insert(index, data)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def pop_row(self, index=-1):
        index = (
            len(self._parsed_data["data"]) - 1
            if index > len(self._parsed_data["data"]) or index < 0
            else index
        )

        if len(self._parsed_data["data"]) != 0:
            popped_row = self._parsed_data["data"].pop(index)
            DataJson()[self.widget_id]["table_data"] = self._parsed_data
            return popped_row

    def get_selected_cell(self, state):
        # logger.debug(
        #     "Selected row",
        #     extra={"selected_row": state[self.widget_id]["selected_row"]},
        # )
        row_index = state[self.widget_id]["selected_row"].get("selectedRow")
        column_name = state[self.widget_id]["selected_row"].get("selectedColumnName")
        column_index = state[self.widget_id]["selected_row"].get("selectedColumn")
        row = state[self.widget_id]["selected_row"].get("selectedRowData")

        if (
            row_index is None
            or column_name is None
            or column_index is None
            or row is None
        ):
            # click table header or clear selection
            return None

        return {
            "column_index": column_index,
            "column_name": column_name,
            "row": row,
            "cell_value": row[column_name],
        }

    def click(self, func):
        route_path = self.get_route_path(Table.Routes.CELL_CLICKED)
        server = self._sly_app.get_server()
        self._click_handled = True

        @server.post(route_path)
        def _click():
            value_dict = self.get_selected_cell(StateJson())
            if value_dict is None:
                return
            datapoint = Table.ClickedDataPoint(**value_dict)
            func(datapoint)

        return _click

    @staticmethod
    def get_html_text_as_button(title="preview"):
        return f"<button>{title}</button>"

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def clear_selection(self):
        StateJson()[self.widget_id]["selected_row"] = {}
        StateJson().send_changes()



from fastapi import FastAPI
from supervisely.app.content import StateJson, DataJson
from supervisely.app.content import get_data_dir
import supervisely.app.fastapi as fastapi
import supervisely.app.widgets as widgets
import supervisely.app.development as development



# coding: utf-8

import os
import time
from collections import defaultdict, OrderedDict
import json

from supervisely.api.module_api import ApiField, ModuleApiBase, ModuleWithStatus, WaitingTimeExceeded
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from supervisely.io.fs import get_file_name, ensure_base_path, get_file_hash
from supervisely.collection.str_enum import StrEnum
from supervisely._utils import batched


class ImageAnnotationToolAction(StrEnum):
    SET_FIGURE = 'figures/setFigure'
    """"""
    NEXT_IMAGE = 'images/nextImage'
    """"""
    PREV_IMAGE = 'images/prevImage'
    """"""
    SET_IMAGE = 'images/setImage'
    """"""
    ZOOM_TO_FIGURE = 'scene/zoomToObject'
    """"""


class ImageAnnotationToolApi(ModuleApiBase):
    def set_figure(self, session_id, figure_id):
        """
        """
        return self._act(session_id, ImageAnnotationToolAction.SET_FIGURE, {ApiField.FIGURE_ID: figure_id})

    def next_image(self, session_id, image_id):
        """
        """
        return self._act(session_id, ImageAnnotationToolAction.NEXT_IMAGE, {ApiField.IMAGE_ID: image_id})

    def prev_image(self, session_id, image_id):
        """
        """
        return self._act(session_id, ImageAnnotationToolAction.PREV_IMAGE, {ApiField.IMAGE_ID: image_id})

    def set_image(self, session_id, image_id):
        """
        """
        return self._act(session_id, ImageAnnotationToolAction.SET_IMAGE, {ApiField.IMAGE_ID: image_id})

    def zoom_to_figure(self, session_id, figure_id, zoom_factor=1):
        """
        """
        return self._act(session_id, ImageAnnotationToolAction.ZOOM_TO_FIGURE,
                         {ApiField.FIGURE_ID: figure_id, ApiField.ZOOM_FACTOR: zoom_factor})

    def _act(self, session_id: int, action: ImageAnnotationToolAction, payload: dict):
        """
        """
        data = {ApiField.SESSION_ID: session_id, ApiField.ACTION: str(action), ApiField.PAYLOAD: payload}
        resp = self._api.post('/annotation-tool.run-action', data)
        return resp.json()



    # {
    #     "sessionId": "940c4ec7-3818-420b-9277-ab3c820babe5",
    #     "action": "scene/setViewport",
    #     "payload": {
    #         "viewport": {
    #             "offsetX": -461, # width
    #             "offsetY": -1228, # height
    #             "zoom": 1.7424000000000024
    #         }
    #     }
    # }

    # {
    #     "sessionId": "940c4ec7-3818-420b-9277-ab3c820babe5",
    #     "action": "scene/zoomToObject",
    #     "payload": {
    #         "figureId": 22129,
    #         "zoomFactor": 1.5
    #     }
    # }


# coding: utf-8
import os


def flag_from_env(s):
    return s.upper() in ['TRUE', 'YES', '1']


def remap_gpu_devices(in_device_ids):
    """
    Working limitation for CUDA
    :param in_device_ids: real GPU devices indexes. e.g.: [3, 4, 7]
    :return: CUDA ordered GPU indexes, e.g.: [0, 1, 2]
    """
    os.environ['CUDA_VISIBLE_DEVICES'] = ','.join(map(str, in_device_ids))
    return list(range(len(in_device_ids)))
